{# -*- engine: jinja -*- #}
<div>
  <table class="legend">
    <tr>
      <th scope="row">Directory:</th>
      <td>{{info.get_directory()}}</td>
    </tr>
    <tr>
      <th scope="row">Date:</th>
      <td>{{info.date}}</td>
    </tr>
  </table>
</div>

<div>
  <table class="coverage">
    <tr>
      <th></th>
      <th scope="col">Exec</th>
      <th scope="col">Total</th>
      <th scope="col">Coverage</th>
    </tr>
    <tr>
      <th scope="row">Lines:</th>
      <td>{{info.lines.exec}}</td>
      <td>{{info.lines.total}}</td>
      <td class="{{info.lines.class}}">{{info.lines.coverage}}%</td>
    </tr>
    <tr>
      <th scope="row">Functions:</th>
      <td>{{info.functions.exec}}</td>
      <td>{{info.functions.total}}</td>
      <td class="{{info.functions.class}}">{{info.functions.coverage}}%</td>
    </tr>
    <tr>
      <th scope="row">Branches:</th>
      <td>{{info.branches.exec}}</td>
      <td>{{info.branches.total}}</td>
      <td class="{{info.branches.class}}">{{info.branches.coverage}}%</td>
    </tr>
    {% if SHOW_DECISION %}
    <tr>
      <th scope="row">Decisions:</th>
      <td>{{info.decisions.exec}}</td>
      <td>{{info.decisions.total}}</td>
      <td class="{{info.decisions.class}}">{{info.decisions.coverage}}%</td>
    </tr>
  {% endif %}
  </table>
</div>
