{# -*- engine: jinja -*- #}
<table class="listOfFunctions">
  <tr>
    <th>Function</th>
    <th>File</th>
    <th>Line</th>
    <th>Call count</th>
  </tr>
{% for entry in all_functions %}
  <tr>
    <td>
      <a href="{{ entry['html_filename'] }}#l{{ entry['line'] }}">{{ entry["name"] }}</a>
    </td>
    <td>
      <a href="{{ entry['html_filename'] }}#l{{ entry['line'] }}">{{ entry["filename"] }}</a>
    </td>
    <td>
      <a href="{{ entry['html_filename'] }}#l{{ entry['line'] }}">{{ entry["line"] }}</a>
    </td>
    <td>
      <a href="{{ entry['html_filename'] }}#l{{ entry['line'] }}">
        {%- if entry["excluded"] %} excluded
        {%- else %}
        {%-  if entry["count"] == 0 %} not called
        {%-  else %} called {{ entry["count"] }} time{% if entry["count"] > 1 %}s{% endif %}
        {%-  endif -%}
        {%- endif -%}
      </a>
    </td>
  </tr>
{% endfor %}
</table>
